#define RANDOM_HYPERGRAPH_DENSITY 100
#define RANDOM_HYPEREDGE_SIZE 2
#define GRAPH_TESTING

//#define PROGRESS_METER

typedef struct hyp_LCAinfo
{
  struct hyp_LCAinfo * child[2];
  struct dt_node * corr_dt;

} hyp_LCAinfo;


typedef int hyperedge;


typedef struct or_tree 
{
  short orientation;
  struct or_tree * child[2]; 


} or_tree;

#ifdef GRAPH_TESTING

typedef struct adjacency
{
  int node;
  struct adjacency * next;

} adjacency;

#endif


typedef struct hypergraph
{
  int size;
  int hyperedges;
  
  int * hyperedge_array;
  int * vertex_array;

  dt_node ** ref_array;

} hypergraph;


typedef struct inverse_hypergraph
{
  int * vertex_array;
  int * hyperedge_array;


} inverse_hypergraph;


//Need the following precompiler macros:

//gives the number of nodes in a hyperedge e from hypergraph g
#define SizeOfHyperedge(e,g) ((g->hyperedge_array[e+1])-(g->hyperedge_array[e]))
//gives the number of edges covering v in inverse hypergraph h
#define NumberOfHyperedgesCoveringVertex(v,h) ((h->vertex_array[v+1])-(h->vertex_array[v]))
//gives the total degree of a hypergraph; the sum of all the hyp. sizes
#define TotalDegree(g) g->hyperedge_array[g->hyperedges]


//NameOfHyperedge(vertex v, int i)
//WeightOfHyperedge(hyperedge e)
//NumberOfHyperedgesInHypergraph(hypergraph * g)
//NumberOfNodesInHypergraph(hypergraph * g)


void FreeOrTree(or_tree * t);
hyp_LCAinfo ** hyp_PreProcessLCA(hypergraph * g, dt_node * dt);
hyp_LCAinfo * hyp_PreProcessLCAForSingleHyperedge(hyperedge e, hypergraph * g, dt_node * dt);
hyp_LCAinfo * hyp_PreProcessLCAForSingleVertexRecursive(dt_node * dt, hyperedge e);
dt_node * FindNextBranchingPoint(dt_node * root, hyperedge e);
dt_node * NewBalancedBinaryTree(hypergraph * g, int levels);
hypergraph * NewRandomHypergraph(int nodes);
inverse_hypergraph * InverseHypergraph(hypergraph * g);
int * hyp_DerivePermutation(dt_node * dt,hypergraph * g);
dt_node *create_decomposition_tree(FILE *hg, // the given hyper-graph
				int ubfactor, // relative imbalance factor
				hypergraph *g // Hypergraph structure  
				   );
dt_node * hyp_LocalImproveIterate(hypergraph * g, dt_node * dt, int ub_factor, int trials, int * min_score);
dt_node * CreateDtFromHypergraph(int ubfactor,hypergraph * g);
hypergraph * CreateHypergraphFromFile(char * filename);
dt_node *do_decomposition(int *node_ids, int nvtxs, int nhedges, int *eptr, int *eind, int ubfactor, hypergraph *g);
dt_node *do_decomposition_SEP(int *node_ids, int nvtxs, int nhedges, int *eptr, int *eind, int ubfactor, hypergraph *g);
or_tree * hyp_OrientOnTheFly(hypergraph * g, dt_node * dt);
or_tree * hyp_OrientWorstOnTheFly(hypergraph * g, dt_node * dt);
or_tree * CreateRandomOrientation(dt_node * dt);
or_tree * CreateZeroOrientation(dt_node * dt);
