// File for basic operations on data structures

#include <stdio.h>
#include <stdlib.h>
#include <math.h>  
#include "ot.h"
#include "hyperorient.h"

//creates a new 2d int array and fills it with zeros
int ** New2dIntArray(int m,int n)
{
  int ** ret;
  int xx_i_xx;
  int j;
  ret = NewArray(int*,m);
  for(xx_i_xx = 0; xx_i_xx < m; xx_i_xx++)
    ret[xx_i_xx] = NewArray(int,n);
  /* clear it */
  for(xx_i_xx = 0; xx_i_xx < m; xx_i_xx++)
    for(j = 0; j < n; j++)
      ret[xx_i_xx][j] = 0;
 
  return ret;
 
 
}

//deallocates the memory for a 2d int array 
void Free2dIntArray(int ** x, int m, int n)
{
 
  int xx_i_xx;
  for(xx_i_xx = 0; xx_i_xx < m; xx_i_xx++)
    free (x[xx_i_xx]);
 
  free(x);
 
}
 

//returns log base 2 of x (closest integer less than)
int log2(int x)
{
  return (int)(log10(x) * 3.32192809488737);
}

//returns 2 to the power of x 
int power2(int x)
{
  int i;
  int r = 1;
  
  for(i=1; i<=x; i++) r = r * 2; 
  
  return r;
}






// Generates a binary tree (allocates space, builds tree, and assigns vertex labels)  
// Tree has input number of levels.  Vertex labels are assigned from 0 to 2^levels - 1.
// Zero levels is considered a single node. Assumes g has exactly 2^levels nodes.
dt_node * GenerateBinaryTree(int levels,        //#levels in tree
			     dt_node * parent, //parent pointer
			     int * counter,   //used to count nodes
			     hypergraph * g)  // graph
{

  dt_node * this_node;
  int counterholder = *counter;

  //Allocate space for this node.
  this_node = New(dt_node);
  this_node->lca_flag = -1;
  this_node->parent = parent;
  this_node->orientation = -1;
  this_node->inner_cut = 0;
  this_node->inner_cut_temp = 0;


  if(levels == 0)
    {
      //Set children to null and assign index as counter
      this_node->child[0] = NULL;
      this_node->child[1] = NULL;
      this_node->v = *counter;
      g->ref_array[(*counter)] = this_node;

      (*counter)++;
    }
  else
    {
      this_node->child[0] = GenerateBinaryTree(levels - 1, this_node, counter, g);
      this_node->child[1] = GenerateBinaryTree(levels - 1, this_node, counter, g);
      this_node->v = NOT_LEAF;
    }

  this_node->subtree_size = (*counter) - counterholder;

  return this_node;
  
  
}



//Wrapper for GenerateBinaryTree. outputs the tree, and also (through parameter)
//a pointer to an array of pointers to the leaves, indexed by graph node 
//assumes g has #nodes equal to a power of 2
dt_node * NewBalancedBinaryTree(hypergraph * g, int levels)
{
  dt_node * root;
  int * counter = New(int);
  *counter = 0;

  root = GenerateBinaryTree(levels, NULL, counter, g);
  
  return root;
}

//Returns a random permutation of a given size
int * RandomPermutation(int size)
{

  int i,j,counter;
  int * pi = NewArray(int,size);
  int * pi_r = NewArray(int,size);
  int picked;

  for(i=0; i<size; i++)
    {
      pi[i] = -1;
      pi_r[i] = -1;
    }

  //pick rand numbers
  for(i=0; i<size; i++)
  {
    picked = RandomNumber(size - i) - 1;
    //printf("executing: %i, number picked: %i\n",i,picked);

    counter = 0;

    for(j=0; counter<=picked; j++)
	if(pi_r[j] == -1)
	  counter++;
  
    pi[i] = j-1;
    pi_r[j-1] = i;

  }

  free(pi_r);
  return pi;

}

//Checks to see if a given permutation is valid
int ValidPermutation(int * pi, int size)
{

  int * a = NewArray(int,size);
  int i;

  for(i=0; i<size; i++)
    a[i] = 0;
  
  
  for(i=0; i<size; i++)
    if(a[pi[i]] == 1)
      return 0;
    else
      a[pi[i]] = 1;



  free(a);
  return 1;
}

//outputs a permutation
void OutputPermutation(int * pi, int size)
{

  int i;
  
  for(i=0; i<size; i++)
      printf("%i: %i\n",i,pi[i]);

}

//derives the normal permutation
int * NormalPermutation(int size)
{

  int * pi = NewArray(int,size);
  int i;


  for(i=0; i<size; i++) pi[i] = i;


  return pi;

}


void FreeDt(dt_node * dt)
{
  int i;
  for(i=0;i<=1;i++) if(dt->child[i] != NULL) FreeDt(dt->child[i]);
  free(dt);
  return;
}

//Procedure to count the number of nodes in an o.t.;  roughly proportional
//to running time.  this code assumes that the orientation of root is fixed.
int SizeOfOt(dt_node * dt, int depth)
{
  int i;
  int size = 0;
    

 
  size = power2(depth-1);
 
  if(dt->v == NOT_LEAF)
    for(i=0;i<=1;i++)
      if(dt->child[i] != NULL)
	size += (SizeOfOt(dt->child[i], depth + 1));   // the +power2(depth) is for this internal node
  
  
  return size;
}





//removes the flagged fields from a dt...  "cleans it up".
void ClearLocalCosts(dt_node * dt)
{
  int i;
  for(i=0;i<=1;i++) if(dt->child[i] != NULL) ClearLocalCosts(dt->child[i]);

  dt->lca_flag = -1;
  dt->orientation = -1;

  return;


}


//Makes a random hypergraph with the input number of nodes the number
//of hyperedges is RANDOM_HYPERGRAPH_DENSITY * the number of nodes.
//the number of nodes in a hypergraph is (RANDOM_HYPEREDGE_SIZE /
//100)* number of nodes
hypergraph * NewRandomHypergraph(int nodes)
{

  hypergraph * ret = New(hypergraph);
  int counter,v,i,j;
  int * temp = NewArray(int, nodes);

  ret->size = nodes;
  ret->hyperedges = (RANDOM_HYPERGRAPH_DENSITY * nodes);

  ret->hyperedge_array = NewArray(int,(RANDOM_HYPERGRAPH_DENSITY * nodes) + 1);
  ret->vertex_array = NewArray(int, RANDOM_HYPERGRAPH_DENSITY * nodes * 
			       RANDOM_HYPEREDGE_SIZE);

  ret->ref_array = NewArray(dt_node*,nodes);

  printf("Creating Random Hypergraph.\nNodes: %i, Hyperedges: %i\n",
	 nodes,ret->hyperedges);


  for(i=0; i<ret->hyperedges; i++)
    {

      ret->hyperedge_array[i] = i*RANDOM_HYPEREDGE_SIZE; 
      //clear temp
      for(j=0; j<nodes; j++)
	temp[j] = 0;
      
      j=0;
      while(j<RANDOM_HYPEREDGE_SIZE)
	{
	  v = RandomNumber(nodes) - 1;
	  if(temp[v] == 0)
	    {
	      temp[v] = 1;
	      ret->vertex_array[i*RANDOM_HYPEREDGE_SIZE + j] = v;
	      j++;
	    }
	}
      
    }

  //set the last entry so that macros will work
  ret->hyperedge_array[i] = i*RANDOM_HYPEREDGE_SIZE; 

  free(temp);
  return ret;
}



inverse_hypergraph * InverseHypergraph(hypergraph * g)
{
  int current_vertex,i;
  int current_hyperedge;
  inverse_hypergraph * ret = New(inverse_hypergraph);
  int * index_array = NewArray(int,g->size+1);

  ret->vertex_array = NewArray(int,g->size+1);
  ret->hyperedge_array = NewArray(int,g->hyperedge_array[g->hyperedges]);
  
  //clear vertex array
  for(i=0;i<=g->size;i++)
      ret->vertex_array[i] = 0;

  //count the degree of each vertex
  for(i=0; i<g->hyperedge_array[g->hyperedges]; i++)
    (ret->vertex_array[g->vertex_array[i]])++;
      
  //fill the index array with the index info
  index_array[0] = 0;
  for(i=1; i<=g->size; i++)
    index_array[i] = index_array[i-1] + ret->vertex_array[i-1];

  //copy the index array into the vertex array
  for(i=0; i<=g->size; i++)
    ret->vertex_array[i] = index_array[i];

  //Now copy the hyperedge names into the hyperedge array
  current_hyperedge = 0;
  for(i=0; i<g->hyperedge_array[g->hyperedges]; i++)
    {
      if(i >= g->hyperedge_array[current_hyperedge + 1])
	current_hyperedge++;

      current_vertex = g->vertex_array[i];
      ret->hyperedge_array[index_array[current_vertex]] = current_hyperedge;
      (index_array[current_vertex])++;

    }

  free(index_array);
  return ret;
}












//Make a given hypergraph into its inverse
inverse_hypergraph * InverseHypergraph_old(hypergraph * g)
{
  inverse_hypergraph * ret = New(inverse_hypergraph);
  int counter;
  int i,j;
  int ** a = NewArray(int*,g->size);
  int * number_of_hyperedges_so_far = NewArray(int,g->size);

  //build matrix;  first index: node, filled from 0 with hyperedges 
  for(i=0; i<g->size; i++)
    {
      number_of_hyperedges_so_far[i] = 0;
      a[i] = NewArray(int,g->hyperedges);
      a[i][0] = -1;
    }
  //Fill matrix
  for(j=0; j<g->hyperedges; j++)
    for(i=g->hyperedge_array[j]; i<g->hyperedge_array[j+1]; i++)
      {
	a[g->vertex_array[i]][number_of_hyperedges_so_far[g->vertex_array[i]]] = j;
	number_of_hyperedges_so_far[g->vertex_array[i]] ++;
	a[g->vertex_array[i]][number_of_hyperedges_so_far[g->vertex_array[i]]] = -1;
      }

  //Build ret from matrix
  ret->vertex_array = NewArray(int,g->size+1);
  ret->hyperedge_array = NewArray(int,TotalDegree(g));

  counter = 0;
  for(i=0; i<g->size; i++)
    {
      ret->vertex_array[i] = counter;
      j=0;
      while(a[i][j] != -1)
	{
	  ret->hyperedge_array[counter] = a[i][j];
	  counter++;
	  j++;
	}
    }

  //put the final index into vertex_array[g->size] for macros
  ret->vertex_array[g->size] = counter;

  //free memory
  free(number_of_hyperedges_so_far);
  for(i=0; i<g->size; i++)
    free(a[i]);

  return ret;
}

void FreeInverseHypergraph(inverse_hypergraph * h)
{


  free(h->vertex_array);
  free(h->hyperedge_array);


}





void FreeHypergraph(hypergraph * g)
{


}


void hyp_OutputHypergraph(hypergraph * g)
{
  int i,j;
  printf("# Nodes:%i, Hyperedges:%i\n",g->size,g->hyperedges);
  for(i=0; i<g->hyperedges; i++)
    {
      for(j=g->hyperedge_array[i]; j<g->hyperedge_array[i+1]; j++)
	printf("%i ",g->vertex_array[j]);
      printf("\n");
    }


}

dt_node * CreateDtFromHypergraph(int ubfactor,hypergraph * g)
{
  dt_node * root;
  int i;
  int *node_ids = malloc(g->size*sizeof(int)); // An array that holds the node numbers
  
  for(i=0; i<g->size; i++)
    node_ids[i]=i;
  
  if(SEPARATE_FLAG)
    root = do_decomposition_SEP(node_ids, g->size, g->hyperedges, 
			    g->hyperedge_array, g->vertex_array, ubfactor, g);
  else
    root = do_decomposition(node_ids, g->size, g->hyperedges, 
			    g->hyperedge_array, g->vertex_array, ubfactor, g);
  
     root->parent = NULL;
  
  free(node_ids);
  
  
  return root;
}

hypergraph * CreateHypergraphFromFile(char * filename)
{
  hypergraph * g = New(hypergraph);
  int i; // Auxiliary variables
  int nvtxs, nhedges; // Number or vertexes and hyperedges
  int *eptr, *eind; // Hyperedges
 
  FILE * fp;

  fp = fopen(filename,"r");
  
  if(fp == NULL)
    {
      printf("File %s Not Found\n",filename);
      return NULL;
    }

 
  extract_from_file(fp, &nvtxs, &nhedges, &eptr, &eind);
 
  g->ref_array = malloc(nvtxs*sizeof(int));
  g->size = nvtxs;
  g->hyperedges=nhedges;
  g->hyperedge_array=eptr;
  g->vertex_array=eind;

  return g;
}








// Notes on arguments:
// ===================
// hg-
//	This is a file pointer to the hyper-graph for which we want to create a decomposition tree.
//	
//
// ubfactor-
//	Properties of ubfactor are characterised in the hmetis documentation.
//	
//
//
// Returned value:
// ===============
// Root of the decomposition tree.




	
